import gzip
import hashlib
import os.path
import tarfile
import tempfile
import io
# noinspection PyCompatibility
import bz2
from moneta.templatetags.moneta import moneta_url

try:
    import lzma
except ImportError:
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    from backports import lzma
import datetime

from django.conf import settings
from django.conf.urls import url
from django.core.urlresolvers import reverse
from django.http.response import Http404, HttpResponse
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.template.loader import render_to_string
from django.utils.timezone import get_current_timezone
from django.utils.translation import ugettext as _

from moneta.archives import ArFile
from moneta.exceptions import InvalidRepositoryException
from moneta.repository.signing import GPGSigner
from moneta.utils import parse_control_data
from moneta.views import get_file, sendpath
from moneta.repositories.base import RepositoryModel
from moneta.repository.models import storage, Repository, Element, ArchiveState


__author__ = 'flanker'

tz = get_current_timezone()


class Aptitude(RepositoryModel):
    verbose_name = _('APT repository for Linux .deb packages')
    storage_uid = 'a97172de-0000-0000-0000-%012d'
    archive_type = 'aptitude'
    index_html = 'repositories/aptitude/index.html'

    def is_file_valid(self, uploaded_file):
        return uploaded_file.name.endswith('.deb')

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data

        ar -x control.tar.gz
        tar -xf control.tar.gz control
        """
        archive_file = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key)
        ar_file = ArFile(element.filename, mode='r', fileobj=archive_file)
        control_file, control_file_name = self.get_subfile(ar_file, 'control.tar.')
        if control_file is None:
            raise InvalidRepositoryException('No control file found in .deb package')
        mode = 'r:*'
        if control_file_name.endswith('.xz') or control_file_name.endswith('.lzma'):
            control_file_content = control_file.read()
            control_file_content_uncompressed = lzma.decompress(control_file_content)
            control_file.close()
            control_file = io.BytesIO(control_file_content_uncompressed)
            mode = 'r'
        tar_file = tarfile.open(name='control', mode=mode, fileobj=control_file)
        control_data = tar_file.extractfile('./control')
        # poulating different informations on the element
        control_data_value = control_data.read().decode('utf-8')
        control_data.close()
        tar_file.close()
        ar_file.close()
        archive_file.close()
        element.extra_data = control_data_value
        parsed_data = parse_control_data(control_data_value)
        element.archive = parsed_data['Package']
        element.version = parsed_data['Version']
        element.official_link = parsed_data.get('Homepage', '')
        element.long_description = parsed_data.get('Description', '')

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        Remove previous versions from target states
        :param element: Element
        :param states: list of ArchiveState
        """
        # remove previous versions from the given states:
        # noinspection PyUnresolvedReferences
        Element.states.through.objects.exclude(element__version=element.version) \
            .filter(archivestate__in=states, element__archive=element.archive).delete()
        super().finish_element(element, states)

    @staticmethod
    def get_subfile(ar_file, prefix='control.tar.'):
        for name in ar_file.getnames():
            if name.startswith(prefix):
                return ar_file.extractfile(name), name
        return None, None

    def file_list(self, element, uid):
        cache_filename = 'filelist_%s' % element.sha256
        key = storage(settings.STORAGE_CACHE).uid_to_key(uid)
        fileobj = storage(settings.STORAGE_CACHE).get_file(key, cache_filename)
        if fileobj is None:
            tmpfile = tempfile.NamedTemporaryFile(dir=settings.TEMP_ROOT)
            archive_file = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key, sub_path='')
            ar_file = ArFile(element.filename, mode='r', fileobj=archive_file)
            data_file, data_file_name = self.get_subfile(ar_file, 'data.tar.')
            mode = 'r:*'
            if data_file_name.endswith('.xz') or data_file_name.endswith('.lzma'):
                data_file_content = data_file.read()
                data_file_content_uncompressed = lzma.decompress(data_file_content)
                data_file.close()
                data_file = io.BytesIO(data_file_content_uncompressed)
                mode = 'r'
            tar_file = tarfile.open(name='data', mode=mode, fileobj=data_file)
            members = tar_file.getmembers()
            members = filter(lambda x: x.isfile(), members)
            names = [x.path[2:] for x in members]
            tar_file.close()
            ar_file.close()
            archive_file.close()
            for name in names:
                tmpfile.write(('%s\n' % name).encode('utf-8'))
            tmpfile.flush()
            tmpfile.seek(0)
            storage(settings.STORAGE_CACHE).store_descriptor(uid, cache_filename, tmpfile)
            tmpfile.close()
        else:
            names = [line.strip().decode() for line in fileobj]
            fileobj.close()
        return names

    def url_list(self):
        """
        Return a list of URL patterns specific to this repository
        Sample recognized urls:
            * /1/pool/repo_apt/stable/[package-1.0.0-amd64.deb]
            * /1/dists/repo_apt/Release[.gpg]
            * /1/dists/repo_apt/Contents-amd64.gz
            * /1/dists/repo_apt/stable/Release.xz
            * /1/dists/repo_apt/stable/binary-amd64/Packages.gz
            * /1/dists/repo_apt/stable/binary-amd64/Release.bz2
            * /1/keys/repo_apt/key.asc

        :return: a patterns as expected by django

        """
        pattern_list = [
            url(r'^(?P<rid>\d+)/force_index/(?P<repo_slug>[\w\-\._]+)/$', self.wrap_view('force_index'), name='force_index'),
        ]
        return pattern_list

    def public_url_list(self):
        """
        Return a list of URL patterns specific to this repository
        Sample recognized urls:
            * /1/pool/repo_apt/stable/[package-1.0.0-amd64.deb]
            * /1/dists/repo_apt/Release[.gpg]
            * /1/dists/repo_apt/Contents-amd64.gz
            * /1/dists/repo_apt/stable/Release.xz
            * /1/dists/repo_apt/stable/binary-amd64/Packages.gz
            * /1/dists/repo_apt/stable/binary-amd64/Release.bz2
            * /1/keys/repo_apt/key.asc

        :return: a patterns as expected by django

        """
        compression = '(?P<compression>|.bz2|.gz|.xz)'
        if lzma is None:
            compression = '(?P<compression>|.bz2|.gz)'
        pattern_list = [
            url(r"^(?P<rid>\d+)/pool/(?P<repo_slug>[\w\-\._]+)/(?P<state_slug>[\w\-\._]+)/(?P<folder>[\w\-\.~_]+)/$",
                self.wrap_view('folder_index'), name="folder_index"),
            url(r"^(?P<rid>\d+)/pool/(?P<repo_slug>[\w\-\._]+)/(?P<state_slug>[\w\-\._]+)/(?P<folder>[\w\-\.~_]+)/"
                r"(?P<filename>[\w\-\.~_ ]+)$",
                self.wrap_view('get_file'), name="get_file"),
            url(r"^(?P<rid>\d+)/dists/(?P<repo_slug>[\w\-\._]+)/(?P<filename>Release|Release.gpg)$",
                self.wrap_view('repo_release'), name="repo_release"),
            url(r"^(?P<rid>\d+)/dists/(?P<repo_slug>[\w\-\._]+)/Contents-(?P<arch>[\w]+)"
                r"%s$" % compression, self.wrap_view('arch_contents'), name="arch_contents"),
            url(r"^(?P<rid>\d+)/dists/(?P<repo_slug>[\w\-\._]+)/(?P<state_slug>[\w\-\._]+)/"
                r"Contents-(?P<arch>[\w\-\.]+)%s$" % compression,
                self.wrap_view('state_contents'), name="state_contents"),
            url(r"^(?P<rid>\d+)/dists/(?P<repo_slug>[\w\-\._]+)/(?P<state_slug>[\w\-\._]+)/binary-(?P<arch>[\w\-\.]+)/"
                r"(?P<filename>Packages|Release)%s$" % compression,
                self.wrap_view('state_files'), name="state_files"),
            url(r"^(?P<rid>\d+)/(?P<slug2>[\w\-\._]+)/(?P<repo_slug>[\w\-\._]+).asc$", self.wrap_view('gpg_key'),
                name="gpg_key"),
            url(r"^(?P<rid>\d+)/$", self.wrap_view('index'), name="index"),
        ]
        return pattern_list

    def extra_index_urls(self, request, repo):
        """ Provides an iterable of tuples (URL, name), to add on the main index
        """
        result = []
        force_index = (reverse('%s:force_index' % self.archive_type, kwargs={'rid': repo.id, 'repo_slug': repo.slug}),
                       _('Index packages'))
        result.append(force_index)
        return result

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def gpg_key(self, request, rid, repo_slug, slug2):
        get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        signature = GPGSigner().export_key()
        return HttpResponse(signature, content_type="text/plain")

    def repo_release(self, request, rid, repo_slug, filename):
        filename = 'dists/%(repo)s/%(filename)s' % {'repo': repo_slug, 'filename': filename}
        return self.index_file(request, rid, filename, 'text/plain')

    def state_files(self, request, rid, repo_slug, state_slug, arch, filename, compression):
        filename = 'dists/%(repo)s/%(state)s/binary-%(arch)s/%(filename)s%(comp)s' % {
            'repo': repo_slug, 'state': state_slug, 'arch': arch, 'filename': filename, 'comp': compression}
        return self.index_file(request, rid, filename, 'text/plain')

    def arch_contents(self, request, rid, repo_slug, arch, compression):
        filename = 'dists/%(repo)s/Contents-%(arch)s%(comp)s' % {'repo': repo_slug, 'arch': arch, 'comp': compression, }
        return self.index_file(request, rid, filename, 'text/plain')

    def state_contents(self, request, rid, repo_slug, state_slug, arch, compression):
        filename = 'dists/%(repo)s/%(state)s/Contents-%(arch)s%(comp)s' % \
                   {'repo': repo_slug, 'arch': arch, 'comp': compression, 'state': state_slug, }
        return self.index_file(request, rid, filename, 'text/plain')

    def index_file(self, request, rid, filename, mimetype):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        uid = self.storage_uid % repo.id
        key = storage(settings.STORAGE_CACHE).uid_to_key(uid)
        return sendpath(settings.STORAGE_CACHE, key, filename, mimetype)

    # noinspection PyUnusedLocal
    def folder_index(self, request, rid, repo_slug, state_slug, folder):
        """
        Return a HttpResponse
        :param request: HttpRequest
        :raise:
        """
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        state = get_object_or_404(ArchiveState, repository=repo, slug=state_slug)
        q = Element.objects.filter(repository__id=rid, archive__startswith=folder, states=state)
        element_query = q.select_related()[0:100]
        element_count = q.count()
        template_values = {'repo': repo, 'state': state_slug, 'element_count': element_count,
                           'elements': element_query, 'folder': folder,
                           'upload_allowed': repo.upload_allowed(request), }
        return render_to_response('repositories/aptitude/folder_index.html', template_values,
                                  RequestContext(request))

    # noinspection PyUnusedLocal
    def get_file(self, request, rid, repo_slug, state_slug, folder, filename):
        """
        Return a HttpResponse
        :param request: HttpRequest
        :raise:
        """
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        q = Element.objects.filter(repository__id=rid, filename=filename)[0:1]
        q = list(q)
        if len(q) == 0:
            raise Http404
        element = q[0]
        return get_file(request, element.id, compression=None, path='', element=element, name=None)

    def index(self, request, rid):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        states = [state for state in ArchiveState.objects.filter(repository=repo).order_by('name')]
        tab_infos = [(states, ArchiveState(name=_('All states'), slug='all-states')), ]
        tab_infos += [([state], state) for state in states]

        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'index_url': reverse(moneta_url(repo, 'index'), kwargs={'rid': repo.id, }),
                           'tab_infos': tab_infos, 'admin_allowed': repo.admin_allowed(request), }
        return render_to_response(self.index_html, template_values, RequestContext(request))

    # noinspection PyUnusedLocal
    def force_index(self, request, rid, repo_slug):
        repo = get_object_or_404(Repository.upload_queryset(request), id=rid, archive_type=self.archive_type)
        self.generate_indexes(repo)
        return HttpResponse(_('Indexes have been successfully rebuilt.'))

    @staticmethod
    def compress_files(open_files: dict, root: str, uid: str) -> list:
        """ Return a list of tuples ((os.path.relpath(filename, root), md5, sha1, sha256, actual_size).
        Also stores the generated files (and original ones)

        :param open_files: dict[filename] = open file descriptor in mode w+b
        :param root:
        :param uid:
        :return:
        """
        hash_controls = []
        for filename, package_file in open_files.items():
            package_file.seek(0)
            gz_filename = filename + '.gz'
            bz2_filename = filename + '.bz2'
            xz_filename = filename + '.xz'
            gz_file = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
            bz2_file = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
            xz_file = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
            bz2_compressor = bz2.BZ2Compressor(9)
            if lzma is not None:
                xz_compressor = lzma.LZMACompressor()
            else:
                xz_compressor = bz2_compressor
            with gzip.GzipFile(gz_filename, mode='wb', compresslevel=9, fileobj=gz_file) as fd_gz:
                data = package_file.read(10240)
                while data:
                    fd_gz.write(data)
                    bz2_file.write(bz2_compressor.compress(data))
                    xz_file.write(xz_compressor.compress(data))
                    data = package_file.read(10240)
            bz2_file.write(bz2_compressor.flush())
            xz_file.write(xz_compressor.flush())
            all_files = [(package_file, filename), (gz_file, gz_filename), (bz2_file, bz2_filename), ]
            if lzma is not None:
                all_files.append((xz_file, xz_filename))
            for obj, filename in all_files:
                obj.flush()
                obj.seek(0)
                data = obj.read(32768)
                md5, sha1, sha256, size = hashlib.md5(), hashlib.sha1(), hashlib.sha256(), 0
                while data:
                    md5.update(data)
                    sha1.update(data)
                    sha256.update(data)
                    size += len(data)
                    data = obj.read(32768)
                hash_controls.append((os.path.relpath(filename, root), md5.hexdigest(), sha1.hexdigest(),
                                      sha256.hexdigest(), size))
                obj.seek(0)
                storage(settings.STORAGE_CACHE).store_descriptor(uid, filename, obj)
                obj.close()
                # build the following files:
        return hash_controls

    def generate_indexes(self, repository, states=None, validity=365):
        default_architectures = {'amd64', }
        uid = self.storage_uid % repository.id
        repo_slug = repository.slug
        root_url = reverse('%s:index' % self.archive_type, kwargs={'rid': repository.id, })
        if repository.is_private:
            root_url = 'authb-%s' % root_url
        if states is None:
            states = list(ArchiveState.objects.filter(repository=repository).order_by('name'))
        states = [state for state in states if
                  Element.objects.filter(repository=repository, states=state).count() > 0]

        all_states_architectures = set()
        all_states = set()
        open_files = {}
        complete_file_list = {}
        root = 'dists/%(repo)s/' % {'repo': repo_slug}
        # list all available architectures (required to add architecture-independent packages to all archs)
        for element in Element.objects.filter(repository=repository):
            control_data = parse_control_data(element.extra_data)
            architecture = control_data.get('Architecture', 'all')
            all_states_architectures.add(architecture)
            # build the following files:
        #   * dists/(group)/(state)/binary-(architecture)/Packages
        #   * dists/(group)/(state)/binary-(architecture)/Release
        # prepare data for:
        #   * dists/(group)/Contents-(architecture)
        if not all_states_architectures or all_states_architectures == {'all'}:
            all_states_architectures = default_architectures
        for state in states:
            state_architectures = set()
            all_states.add(state.name)
            for element in Element.objects.filter(repository=repository, states=state).order_by('filename'):
                control_data = parse_control_data(element.extra_data)
                architecture = control_data.get('Architecture', 'all')
                section = control_data.get('Section', 'contrib')
                package_file_list = ["%- 100s%s\n" % (x, section) for x in self.file_list(element, uid)]
                if architecture == 'all':
                    elt_architectures = default_architectures
                else:
                    elt_architectures = {architecture, }
                state_architectures |= elt_architectures
                for architecture in elt_architectures:
                    complete_file_list.setdefault(architecture, [])
                    complete_file_list[architecture] += package_file_list

                    filename = 'dists/%(repo)s/%(state)s/binary-%(architecture)s/Packages' % {
                        'repo': repo_slug, 'state': state.name, 'architecture': architecture, }
                    if filename not in open_files:
                        open_files[filename] = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
                    package_file = open_files[filename]
                    package_file.write(element.extra_data.encode('utf-8'))
                    for key, attr in (('MD5sum', 'md5'), ('SHA1', 'sha1'), ('SHA256', 'sha256'),
                                      ('Size', 'filesize')):
                        if key not in control_data:
                            package_file.write("{0}: {1}\n".format(key, getattr(element, attr)).encode('utf-8'))
                    package_url = reverse('%s:get_file' % self.archive_type,
                                          kwargs={'rid': repository.id, 'repo_slug': repo_slug,
                                                  'filename': element.filename, 'state_slug': state.slug,
                                                  'folder': element.filename[0:1], })
                    package_url = os.path.relpath(package_url, root_url)
                    package_file.write("Filename: {0}\n".format(package_url).encode('utf-8'))
                    package_file.write("\n".encode('utf-8'))
            if len(state_architectures) == 0:
                state_architectures = default_architectures
                # we process elements
            for architecture in state_architectures:
                filename = 'dists/%(repo)s/%(state)s/binary-%(architecture)s/Release' % {
                    'repo': repo_slug, 'state': state.slug, 'architecture': architecture,
                }
                open_files[filename] = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
                content = render_to_string('repositories/aptitude/architecture_release.txt',
                                           {'architecture': architecture, 'repository': repository, 'state': state, })
                open_files[filename].write(content.encode('utf-8'))
            # build the following files:
        #   * dists/(group)/Contents-(architecture)
        for architecture, file_list in complete_file_list.items():
            file_list.sort()
            filename = 'dists/%(repo)s/Contents-%(architecture)s' % {'repo': repo_slug,
                                                                     'architecture': architecture, }
            open_files[filename] = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
            open_files[filename].write(render_to_string('repositories/aptitude/contents.txt').encode('utf-8'))
            for info in file_list:
                open_files[filename].write(info.encode('utf-8'))
            # build the following files:
        #   * dists/(group)/Contents-(architecture).gz/.bz2/.xz
        #   * dists/(group)/(state)/binary-(architecture)/Packages.gz/.bz2/.xz
        #   * dists/(group)/(state)/binary-(architecture)/Release.gz/.bz2/.xz
        # store all files in the cache
        hash_controls = self.compress_files(open_files, root, uid)
        #   * dists/(group)/Release
        # store all files in the cache
        release_file = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
        now = datetime.datetime.now(tz)
        now_str = now.strftime('%a, %d %b %Y %H:%M:%S %z')  # 'Mon, 29 Nov 2010 08:12:51 UTC'
        until = (now + datetime.timedelta(validity)).strftime('%a, %d %b %Y %H:%M:%S %z')
        content = render_to_string('repositories/aptitude/state_release.txt',
                                   {'architectures': all_states_architectures, 'until': until,
                                    'states': all_states, 'repository': repository, 'date': now_str})
        release_file.write(content.encode('utf-8'))
        for hash_value, index in (('MD5Sum', 1), ('SHA1', 2), ('SHA256', 3)):
            release_file.write("{0}:\n".format(hash_value).encode('utf-8'))
            for line in hash_controls:
                release_file.write((" %s % 8d %s\n" % (line[index], line[4], line[0])).encode('utf-8'))
        release_file.flush()
        release_file.seek(0)
        filename = 'dists/%(repo)s/Release' % {'repo': repo_slug, }
        storage(settings.STORAGE_CACHE).store_descriptor(uid, filename, release_file)
        # build the following files:
        #   * dists/(group)/Release.gpg
        # store all files in the cache
        release_file.seek(0)
        signature = GPGSigner().sign_file(release_file)
        release_file.close()

        gpg_file = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
        gpg_file.write(signature.encode('utf-8'))
        gpg_file.flush()
        gpg_file.seek(0)
        filename = 'dists/%(repo)s/Release.gpg' % {'repo': repo_slug, }
        storage(settings.STORAGE_CACHE).store_descriptor(uid, filename, gpg_file)
        gpg_file.close()



# coding=utf-8
import base64
from distutils.version import LooseVersion
import hashlib
import mimetypes
import os
import tarfile
import tempfile
import zipfile
import time

from django import forms
from django.conf import settings
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.core.files.uploadedfile import UploadedFile
from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage
from django.core.urlresolvers import reverse
from django.core.validators import RegexValidator
from django.db.models import Count
from django.http import HttpResponseRedirect, Http404, StreamingHttpResponse, HttpResponse, HttpRequest
from django.views.decorators.cache import never_cache

from django.views.decorators.csrf import csrf_exempt
from djangofloor.views import send_file

from moneta.repository.signing import GPG
from moneta.exceptions import InvalidRepositoryException
from moneta.repository.forms import get_repository_form, RepositoryUpdateForm
from moneta.utils import read_file_in_chunks
from moneta.repository.models import Repository, ArchiveState, Element, storage, ElementSignature

__author__ = 'flanker'

from django.shortcuts import render_to_response, get_object_or_404
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _


@never_cache
def index(request: HttpRequest):
    repositories = Repository.index_queryset(request).annotate(package_count=Count('element'))
    if not request.user.has_perm('repository.add_repository'):
        form = None
    elif request.method == 'POST':
        form = get_repository_form()(request.POST)
        if form.is_valid():
            author = None if request.user.is_anonymous() else request.user
            repo = Repository(author=author, name=form.cleaned_data['name'], on_index=form.cleaned_data['on_index'],
                              archive_type=form.cleaned_data['archive_type'],
                              is_private=form.cleaned_data['is_private'])
            repo.save()
            for group in form.cleaned_data['admin_group']:
                repo.admin_group.add(group)
            for state in set(form.cleaned_data['states'].split()):
                ArchiveState(repository=repo, name=state, author=author).save()
            messages.info(request, _('Your new repository has been created.'))
            return HttpResponseRedirect(reverse('moneta:index'))
    else:
        form = get_repository_form()()
    # compute repos with admin rights
    upload_ids = {x.id for x in Repository.upload_queryset(request)}
    admin_ids = {x.id for x in Repository.admin_queryset(request)}
    template_values = {'repositories': repositories, 'form': form, 'request': request,
                       'upload_ids': upload_ids, 'admin_ids': admin_ids, }
    return render_to_response('moneta/index.html', template_values, RequestContext(request))


def delete_repository(request: HttpRequest, rid):
    from moneta.repository.forms import DeleteRepositoryForm

    repo = get_object_or_404(Repository.upload_queryset(request), id=rid)
    if request.method == 'POST':
        form = DeleteRepositoryForm(request.POST)
        if form.is_valid():
            for element in Element.objects.filter(repository=repo):
                element.delete()
            repo.delete()
            messages.warning(request, _('The repository %(repo)s has been deleted.') % {'repo': repo.name})
            return HttpResponseRedirect(reverse('moneta:index'))
    else:
        form = DeleteRepositoryForm()
    template_values = {'form': form, 'repo': repo}
    return render_to_response('moneta/delete_repo.html', template_values, RequestContext(request))


@never_cache
def public_check(request: HttpRequest):
    messages.success(request, _('You can access to this page.'))
    return render_to_response('moneta/empty.html', RequestContext(request))


@never_cache
def private_check(request: HttpRequest):
    if request.user.is_anonymous():
        messages.error(request, _('You are not authenticated.'))
    else:
        messages.success(request, _('You can access to this page and you are authenticated.'))
    return render_to_response('moneta/empty.html', RequestContext(request))


@never_cache
def check(request: HttpRequest):
    s_h = settings.SECURE_PROXY_SSL_HEADER
    a_h = settings.AUTHENTICATION_HEADER
    # noinspection PyArgumentList
    gpg_valid = False
    for key in GPG.list_keys(False):
        if key['keyid'] == settings.GNUPG_KEYID:
            gpg_valid = True
    import moneta.core.defaults
    import moneta.core.settings as real_settings

    default_conf_path = moneta.core.defaults.__file__
    if default_conf_path.endswith('.pyc'):
        default_conf_path = default_conf_path[:-1]

    template_values = {
        'media_root': settings.MEDIA_ROOT, 'media_url': settings.MEDIA_URL,
        'static_root': settings.STATIC_ROOT, 'static_url': settings.STATIC_URL,
        'debug': settings.DEBUG, 'use_x_forwared_host': settings.USE_X_FORWARDED_HOST,
        'secure_proxy_ssl_header_name': s_h[0][5:],
        'secure_proxy_ssl_header_value': s_h[1],
        'authentication_header': a_h,
        'has_authentication_header': request.META.get(a_h, '') == request.user.username,
        'has_secure_proxy_ssl_header': request.META.get(s_h[0], '') == s_h[1],
        'host': request.get_host().rpartition(':')[0],
        'has_allowed_host': request.get_host() in settings.ALLOWED_HOSTS,
        'gpg_valid': gpg_valid, 'gpg_available': GPG.list_keys(False),
        'conf_path': getattr(settings, 'CONF_PATH', None), 'conf_is_set': settings.CONF_IS_SET,
        'default_conf_path': default_conf_path, 'settings': real_settings.SETTINGS_VARIABLE
    }

    return render_to_response('moneta/help.html', template_values, RequestContext(request))


@never_cache
def modify_repository(request: HttpRequest, rid):
    repo = get_object_or_404(Repository.upload_queryset(request), id=rid)
    author = None if request.user.is_anonymous() else request.user

    if request.method == 'POST':
        form = RepositoryUpdateForm(request.POST)
        if form.is_valid():
            new_state_names = set(form.cleaned_data['states'].split())
            old_state_names = {x.name for x in repo.archivestate_set.all()}
            for name in new_state_names - old_state_names:  # new states
                ArchiveState(name=name, repository=repo, author=author).save()
            repo.on_index = form.cleaned_data['on_index']
            repo.is_private = form.cleaned_data['is_private']
            repo.save()
            repo.admin_group.clear()
            for group in form.cleaned_data['admin_group']:
                repo.admin_group.add(group)
            # repo.reader_group.clear()
            # for group in form.cleaned_data['reader_group']:
            #     repo.reader_group.add(group)
            removed_states = ArchiveState.objects.filter(name__in=old_state_names - new_state_names, repository=repo)
            # noinspection PyUnresolvedReferences
            Element.states.through.objects.filter(archivestate__in=removed_states).delete()
            removed_states.delete()
            messages.info(request, _('The repository %(repo)s has been modified.') % {'repo': repo.name})
            return HttpResponseRedirect(reverse('moneta:modify_repository', kwargs={'rid': rid, }))
    else:
        form = RepositoryUpdateForm(initial={'on_index': repo.on_index, 'is_private': repo.is_private,
                                             'reader_group': list(repo.reader_group.all()),
                                             'states': ' '.join([x.name for x in repo.archivestate_set.all()]),
                                             'admin_group': list(repo.admin_group.all())})
    template_values = {'form': form, 'repo': repo, 'upload_allowed': repo.upload_allowed(request)}
    return render_to_response('moneta/modify_repo.html', template_values, RequestContext(request))


@never_cache
def search_package(request: HttpRequest, rid):
    repo = get_object_or_404(Repository.reader_queryset(request), id=rid)
    repo_states = list(ArchiveState.objects.filter(repository=repo))

    class ElementSearchForm(forms.Form):
        search = forms.CharField(label=_('Search'), help_text=_('Type your search'), max_length=200, required=False)
        states = forms.ModelMultipleChoiceField(label=_('Selected states'), required=False,
                                                initial=repo_states,
                                                queryset=ArchiveState.objects.filter(repository=repo), )
        states.help_text = ''

    if request.GET.get('search') is not None:
        form = ElementSearchForm(request.GET)
    else:
        form = ElementSearchForm()
    query = Element.objects.filter(repository=repo)
    search_pattern = ''
    if form.is_valid():
        search_pattern = form.cleaned_data['search']
        if search_pattern:
            query = query.filter(full_name__icontains=search_pattern)
        if form.cleaned_data['states']:
            query = query.filter(states__in=form.cleaned_data['states']).distinct()
    else:
        query = query.filter(states__in=repo_states).distinct()

    paginator = Paginator(query.order_by('full_name'), 25)
    page = request.GET.get('page')

    try:
        elements = paginator.page(page)
    except PageNotAnInteger:
        # If page is not an integer, deliver first page.
        elements = paginator.page(1)
    except EmptyPage:
        # If page is out of range (e.g. 9999), deliver last page of results.
        elements = paginator.page(paginator.num_pages)
    template_values = {'elements': elements, 'repo': repo, 'pattern': search_pattern, 'form': form,
                       'upload_allowed': repo.upload_allowed(request)}
    return render_to_response('moneta/search_repo.html', template_values, RequestContext(request))


def delete_element(request: HttpRequest, rid, eid):
    from moneta.repository.forms import DeleteRepositoryForm

    repo = get_object_or_404(Repository.upload_queryset(request), id=rid)
    element = get_object_or_404(Element.objects.filter(repository=repo, id=eid))
    if request.method == 'POST':
        form = DeleteRepositoryForm(request.POST)
        if form.is_valid():
            element.delete()
            messages.warning(request, _('The package %(repo)s has been deleted.') % {'repo': element.full_name})
            return HttpResponseRedirect(reverse('moneta:index'))
    else:
        form = DeleteRepositoryForm()
    template_values = {'form': form, 'repo': repo, 'element': element}
    return render_to_response('moneta/delete_element.html', template_values, RequestContext(request))


def generic_add_element(request: HttpRequest, repo, uploaded_file, state_names, archive=None, name=None, version=None):
    """ Generic upload (both form-based and raw POST-based methods)
    :param repo: Repository
    :param uploaded_file: UploadedFile
    :param state_names: iterable of ArchiveState.name
    :return: successfully added Element
    """
    sha256 = hashlib.sha256()
    data = uploaded_file.file.read(4096)
    while data:
        sha256.update(data)
        data = uploaded_file.file.read(4096)
    sha256_sum = sha256.hexdigest()
    uploaded_file.file.seek(0)
    if uploaded_file.name:
        filename = os.path.basename(uploaded_file.name)
        elements = list(Element.objects.filter(repository=repo, filename=filename)[0:1])
    elif name:
        name = os.path.basename(name)
        elements = list(Element.objects.filter(repository=repo, name=name)[0:1])
    else:
        elements = list(Element.objects.filter(repository=repo, sha256=sha256_sum)[0:1])
    if elements:
        element = elements[0]
    else:
        user = None if request.user.is_anonymous() else request.user
        element = Element(repository=repo, author=user)
    if archive:
        element.archive = archive
    if name:
        element.name = os.path.basename(name)
    if version:
        element.version = version
    element.archive_file = uploaded_file
    # form.cleaned_data['package'].file.name = form.cleaned_data['package'].name
    element.save()
    cls = element.repository.get_model()
    states = list(ArchiveState.objects.filter(repository=repo, slug__in=state_names))
    cls.finish_element(element, states)
    return element


@never_cache
def add_element(request: HttpRequest, rid):
    repo = get_object_or_404(Repository.upload_queryset(request), id=rid)

    class ElementForm(forms.Form):
        package = forms.FileField(label=_('Package'))
        states = forms.ModelMultipleChoiceField(repo.archivestate_set.all(), label=_('States'))

        def clean(self):
            data_ = super().clean()
            if 'package' in self.cleaned_data:
                if not repo.get_model().is_file_valid(self.cleaned_data['package']):
                    raise ValidationError(_('This repository is unable to handle this file.'))
            return data_

    if request.method == 'POST':
        form = ElementForm(request.POST, files=request.FILES)
        if form.is_valid():
            try:
                element = generic_add_element(request, repo, form.cleaned_data['package'],
                                              [x.name for x in form.cleaned_data['states']])
                messages.info(request, _('The package %(n)s has been successfully uploaded.') % {'n': element.filename})
            except InvalidRepositoryException as e:
                messages.error(request, _('Unable to add the package to this repository: %(msg)s.') % {'msg': str(e)})
            return HttpResponseRedirect(reverse('moneta:add_element', kwargs={'rid': rid}))
    else:
        form = ElementForm()
    template_values = {'form': form, 'repo': repo, 'upload_allowed': repo.upload_allowed(request)}
    return render_to_response('moneta/add_package.html', template_values, RequestContext(request))


@csrf_exempt
def add_element_signature(request: HttpRequest, rid):
    from moneta.repository.forms import SignatureForm

    if request.method != 'POST':
        return render_to_response('moneta/not_allowed.html', status=405)
    form = SignatureForm(request.GET)
    if not form.is_valid():
        return render_to_response('moneta/not_allowed.html', status=405)
    signature = base64.b64encode(request.read(16384))
    sha256 = form.cleaned_data['sha256']
    method = form.cleaned_data['method']
    user = None if request.user.is_anonymous() else request.user
    element = get_object_or_404(Element.reader_queryset(request), repository__id=rid, sha256=sha256, author=user)
    ElementSignature(element=element, signature=signature, method=method).save()
    return HttpResponse(_('This signature has been added to %(filename)s') % {'filename': element.filename})


@csrf_exempt
def add_element_post(request: HttpRequest, rid):
    try:
        repo = Repository.upload_queryset(request).get(id=rid)
    except Repository.DoesNotExist:
        return HttpResponse(_('You cannot upload new packages to this repository'), status=403)
    if request.method != 'POST':
        return HttpResponse(_('Method not allowed'), status=405)
    validators = [RegexValidator(r'[\w\.\-\(\)/]+')]

    class ElementForm(forms.Form):
        filename = forms.CharField(label=_('Package'), max_length=255, validators=validators)
        states = forms.MultipleChoiceField([(x.name, x.name) for x in repo.archivestate_set.all()], label=_('States'))
        name = forms.CharField(label=_('Name'), max_length=255, validators=validators, required=False)
        archive = forms.CharField(label=_('Archive'), max_length=255, validators=validators, required=False)
        version = forms.CharField(label=_('Version'), max_length=255, validators=validators, required=False)

    form = ElementForm(request.GET)
    if not form.is_valid():
        return render_to_response('moneta/not_allowed.html', status=405)

    tmp_file = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
    c = False
    chunk = request.read(32768)
    while chunk:
        tmp_file.write(chunk)
        c = True
        chunk = request.read(32768)
    tmp_file.flush()
    tmp_file.seek(0)
    if not c:
        return HttpResponse(_('Empty file. You must POST a valid file.\n'), status=400)
    uploaded_file = UploadedFile(name=form.cleaned_data['filename'], file=tmp_file)
    try:
        element = generic_add_element(request, repo, uploaded_file, form.cleaned_data['states'],
                                      name=form.cleaned_data.get('name'), archive=form.cleaned_data.get('archive'),
                                      version=form.cleaned_data.get('version'), )
    except InvalidRepositoryException as e:
        return HttpResponse(str(e), status=400)
    finally:
        tmp_file.close()
    template_values = {'repo': repo, 'element': element}
    return HttpResponse(_('Package %(element)s successfully added to repository %(repo)s.\n') % template_values)


@never_cache
def show_file(request: HttpRequest, eid):
    q = Element.reader_queryset(request).filter(id=eid).select_related()[0:1]
    elements = list(q)
    if len(elements) == 0:
        raise Http404
    element = elements[0]
    template_values = {'element': element, 'repo': element.repository,
                       'upload_allowed': element.repository.upload_allowed(request)}
    return render_to_response('moneta/show_package.html', template_values, RequestContext(request))


def get_checksum(request: HttpRequest, eid, value):
    element = get_object_or_404(Element.reader_queryset(request), id=eid)
    value = getattr(element, value)
    return HttpResponse('%s  %s\n' % (value, element.filename), content_type='text/plain')


def get_signature(request: HttpRequest, eid, sid):
    element = get_object_or_404(Element.reader_queryset(request), id=eid)
    signature = get_object_or_404(ElementSignature, element=element, id=sid)
    value = base64.b64decode(signature.signature)
    mimetype = 'application/pgp-signature' if signature.method == signature.GPG else 'application/x509'
    return HttpResponse(value, content_type=mimetype)


def get_checksum_p(request: HttpRequest, eid, value):
    return get_checksum(request, eid, value)


def get_file_p(request: HttpRequest, eid, compression=None, path='', element=None, name=None):
    return get_file(request, eid, compression=compression, path=path, element=element, name=name)


def get_signature_p(request: HttpRequest, eid, sid):
    return get_signature(request, eid, sid)


def get_file(request: HttpRequest, eid: int, compression: str=None, path: str='', element: Element=None, name: str=None):
    """
    Send file to the client as a HttpResponse
    Multiple combinations:
        * case 1) if path != '' => send a file inside a compressed archive
        * case 2) elif compression is not None => required uncompressed archive to compress it to the new format
        * case 3) else => require original file

    :param request:
    :param eid:
    :param compression:
    :param path:
    :param element: avoid an extra DB query to fetch 
    :param name:
    :return:
    """
    # noinspection PyUnusedLocal
    name = name
    if element is None:
        element = get_object_or_404(Element.reader_queryset(request).select_related(), id=eid)
    arc_storage, arc_key, arc_path = None, None, None
    mimetype = 'application/octet-stream'
    if element.uncompressed_key and path:  # case 1
        path = os.path.normpath(path)
        if path.startswith('../'):
            raise Http404
    elif element.uncompressed_key and compression is not None:  # case 2
        arc_storage, arc_key, arc_path = storage(settings.STORAGE_UNCOMPRESSED), element.uncompressed_key, path
    elif element.archive_key:  # case 2 or 3
        if compression is not None:  # case 2
            arc_storage, arc_key, arc_path = storage(settings.STORAGE_ARCHIVE), element.archive_key, ''
    else:
        raise Http404
    if arc_storage is not None:
        temp_file = tempfile.TemporaryFile(mode='w+b', dir=settings.TEMP_ROOT)
        comp_file = None
        ext = ''
        if compression == 'zip':
            mimetype = 'application/zip'
            ext = '.zip'
            comp_file = zipfile.ZipFile(temp_file, 'w', zipfile.ZIP_DEFLATED)
        elif compression == 'tgz':
            mimetype = 'application/x-tar'
            ext = '.tgz'
            comp_file = tarfile.open(None, 'w:gz', fileobj=temp_file)
            comp_file.write = comp_file.addfile
        elif compression == 'tbz':
            mimetype = 'application/x-tar'
            ext = '.tbz'
            comp_file = tarfile.open(None, 'w:bz2', fileobj=temp_file)
            comp_file.write = comp_file.addfile
        reldir = None
        for root, dirs, files in arc_storage.walk(arc_key, arc_path):
            if reldir is None:
                reldir = root
            for name in files:
                fullname = os.path.join(root, name)
                fileobj = arc_storage.get_file(arc_key, fullname)
                arcname = os.path.relpath(fullname, reldir)
                tarinfo = tarfile.TarInfo(arcname)
                tarinfo.size = arc_storage.get_size(arc_key, fullname)
                comp_file.write(tarinfo, fileobj)
        comp_file.close()
        temp_file.seek(0)
        fileobj = temp_file
        filename = os.path.basename(element.filename) + ext
    elif path:
        mimetype = mimetypes.guess_type(path)[0]
        if mimetype is None:
            mimetype = 'application/octet-stream'
        return sendpath(settings.STORAGE_UNCOMPRESSED, element.uncompressed_key, path, mimetype)
    else:
        return sendpath(settings.STORAGE_ARCHIVE, element.archive_key, '', element.mimetype)
    response = StreamingHttpResponse(read_file_in_chunks(fileobj), content_type=mimetype)
    if mimetype[0:4] != 'text' and mimetype[0:5] != 'image':
        response['Content-Disposition'] = 'attachment; filename={0}'.format(filename)
    return response


def sendpath(storage_type, key, path, mimetype):
    storage_obj = storage(storage_type)
    filesize = storage_obj.get_size(key, path)
    full_path = storage_obj.get_path(key, path)
    if full_path:
        return send_file(full_path)
    fileobj = storage_obj.get_file(key, path)
    if fileobj is None:
        raise Http404
    response = StreamingHttpResponse(read_file_in_chunks(fileobj), content_type=mimetype)
    if mimetype[0:4] != 'text' and mimetype[0:5] != 'image':
        response['Content-Disposition'] = 'attachment; filename={0}'.format(os.path.basename(path))
    response['Content-Length'] = filesize
    return response


def compare_states(request: HttpRequest, rid):
    repo = get_object_or_404(Repository.reader_queryset(request), id=rid)
    states = ArchiveState.objects.filter(repository=repo)
    operators = {'<': lambda x, y: x < y, '≤': lambda x, y: x <= y, '=': lambda x, y: x == y,
                 '≥': lambda x, y: x >= y, '>': lambda x, y: x > y}

    class CompareForm(forms.Form):
        state_left = forms.ModelChoiceField(states, label=_('Left state'))
        operator = forms.ChoiceField([(x, x) for x in operators], label=_('Assert operator'))
        state_right = forms.ModelChoiceField(states, label=_('Right state'))

    template_values = {'repo': repo, 'upload_allowed': repo.upload_allowed(request)}
    if request.method == 'POST':
        form = CompareForm(request.POST)
        if form.is_valid():
            state_left = form.cleaned_data['state_left']
            state_right = form.cleaned_data['state_right']
            operator = operators[form.cleaned_data['operator']]
            left_elements = Element.objects.filter(repository=repo, states=state_left)
            right_elements = Element.objects.filter(repository=repo, states=state_right)
            left_dict = {}
            right_dict = {}
            for element in left_elements:
                left_dict.setdefault(element.archive, (LooseVersion(element.version), []))[1].append(element)
            for element in right_elements:
                right_dict.setdefault(element.archive, (LooseVersion(element.version), []))[1].append(element)
            only_left, only_right = {}, {}
            left_keys, right_keys = set(left_dict.keys()), set(right_dict.keys())
            for key in left_keys - right_keys:
                only_left[key] = left_dict[key]
            for key in right_keys - left_keys:
                only_right[key] = right_dict[key]
            invalid_values = []
            for key in left_keys.intersection(right_keys):
                if not operator(left_dict[key][0], right_dict[key][0]):
                    invalid_values.append((key, left_dict[key], right_dict[key]))
            invalid_values.sort(key=lambda x: x[0])
            template_values.update({'only_left': only_left, 'only_right': only_right, 'invalid_values': invalid_values,
                                    'state_left': state_left, 'state_right': state_right})
            template_values['all_valid'] = bool(not only_left and not only_right and not invalid_values)
    else:
        form = CompareForm()
    template_values['form'] = form
    return render_to_response('moneta/compare_states.html', template_values, RequestContext(request))


@csrf_exempt
def test_upload(request: HttpRequest):
    start = time.time()
    size = 0.
    chunk = request.read(32768)
    while chunk:
        size += len(chunk)
        chunk = request.read(32768)
    end = time.time()
    ans = '%d s : %g o/s' % ((end - start), (size / (end - start)))
    return HttpResponse(ans.encode('utf-8'))



# coding=utf-8

from django.conf.urls import include, url


def get_patterns():
    from moneta.repositories.base import RepositoryModelsClasses
    pattern_list_ = []
    for name, model in RepositoryModelsClasses().get_models().items():
        pattern_list_.append(url(r'^p/%s/' % name, include(model.public_urls, namespace=name)))
    pattern_list_.append(url(r'^p/$', 'moneta.repository.views.public_check'))
    for name, model in RepositoryModelsClasses().get_models().items():
        pattern_list_.append(url(r'^a/%s/' % name, include(model.urls, namespace='auth-' + name)))
        pattern_list_.append(url(r'^b/%s/' % name, include(model.public_urls, namespace='authb-' + name)))
    pattern_list_.append(url(r'^a/$', 'moneta.repository.views.private_check'))
    pattern_list_.append(url(r'^b/$', 'moneta.repository.views.private_check'))
    return pattern_list_


name_set = set()
urlpatterns = get_patterns()



# coding=utf-8
from allauth.account.models import EmailConfirmation, EmailAddress
from allauth.socialaccount.models import SocialApp, SocialToken
from allauth.socialaccount.models import SocialAccount
from django.contrib.sites.models import Site
from django.contrib import admin

from moneta.repository.models import Element, Repository


class ElementAdmin(admin.ModelAdmin):
    pass

admin.site.unregister(Site)
admin.site.unregister(SocialToken)
admin.site.unregister(SocialAccount)
admin.site.unregister(SocialApp)
admin.site.unregister(EmailAddress)
admin.site.unregister(EmailConfirmation)

admin.site.register(Element, ElementAdmin)
admin.site.register(Repository)



# -*- coding: utf-8 -*-
"""Emulate a Jetbrains repository.

"""
from django.conf.urls import url
from django.core.urlresolvers import reverse
from django.http import HttpRequest
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.utils.translation import ugettext as _

from moneta.repositories.base import RepositoryModel
from moneta.repository.models import Repository, ArchiveState, Element
from moneta.templatetags.moneta import moneta_url

__author__ = 'Matthieu Gallet'


class Jetbrains(RepositoryModel):
    verbose_name = _('Jetbrains repository for IDEs')
    storage_uid = 'jetbrain-0000-0000-0000-%012d'
    archive_type = 'jetbrains'
    index_html = 'repositories/jetbrains/index.html'

    def public_url_list(self):
        return [
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/s/(?P<state_slug>[\w\-\._]+)/updatePlugins.xml$", self.wrap_view('plugin_index'),
                name='plugin_index'),
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/a/updatePlugins.xml$", self.wrap_view('plugin_index'), name='plugin_index'),
            url(r"^(?P<rid>\d+)/$", self.wrap_view('index'), name="index"),
        ]

    def plugin_index(self, request: HttpRequest, rid, repo_slug, state_slug=None):
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        return render_to_response('repositories/jetbrains/updatePlugins.xml', {'elements': base_query}, RequestContext(request), content_type='application/xml')

    def index(self, request, rid):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        states = [state for state in ArchiveState.objects.filter(repository=repo).order_by('name')]
        tab_infos = [(reverse('jetbrains:plugin_index', kwargs={'rid': repo.id, 'repo_slug': repo.slug}),
                      ArchiveState(name=_('All states'), slug='all-states'), states), ]
        tab_infos += [(reverse('jetbrains:plugin_index', kwargs={'rid': repo.id, 'repo_slug': repo.slug, 'state_slug': state.slug}), state, [state])
                      for state in states]

        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'index_url': reverse(moneta_url(repo, 'index'), kwargs={'rid': repo.id, }),
                           'tab_infos': tab_infos, 'admin_allowed': repo.admin_allowed(request), }
        return render_to_response(self.index_html, template_values, RequestContext(request))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# coding=utf-8
from bootstrap3.templatetags.bootstrap3 import get_pagination_context
from django import template
from django.core.urlresolvers import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

register = template.Library()


@register.filter
def extra_repo_urls(repo, request):
    return repo.get_model().extra_index_urls(request, repo)


@register.filter
def moneta_url(repo, view_name='index'):
    if repo.is_private:
        return 'authb-%s:%s' % (repo.archive_type, view_name)
    return '%s:%s' % (repo.archive_type, view_name)


@register.filter
def checksum(element, value):
    if element.repository.is_private:
        return reverse('moneta:get_checksum', kwargs={'eid': element.id, 'value': value})
    return reverse('moneta:get_checksum_p', kwargs={'eid': element.id, 'value': value})


@register.filter
def direct_link(element):
    if element.repository.is_private:
        return reverse('moneta:get_file', kwargs={'eid': element.id, 'name': element.filename})
    return reverse('moneta:get_file_p', kwargs={'eid': element.id, 'name': element.filename})


@register.filter
def human_join(list_of_elements):
    if not list_of_elements:
        return ''
    elif len(list_of_elements) == 1:
        return list_of_elements[0]
    result = _(', ').join([str(x) for x in list_of_elements[:-1]])
    result += _(' and ') + str(list_of_elements[-1])
    return result


@register.filter
def signature(signature_, element=None):
    if element is None:
        element = signature_.element
    if element.repository.is_private:
        return reverse('moneta:get_signature', kwargs={'eid': element.id, 'sid': signature_.id})
    return reverse('moneta:get_signature_p', kwargs={'eid': element.id, 'sid': signature_.id})


@register.filter
def auth_moneta_url(repo, view_name='index'):
    return 'auth-%s:%s' % (repo.archive_type, view_name)


@register.inclusion_tag('bootstrap3/pagination.html')
def bootstrap_pagination_extra(page, search=None):
    """
    Render pagination for a page

    **Tag name**::

        bootstrap_pagination

    **Parameters**:

        :page:
        :kwargs:

    **usage**::

        {% bootstrap_pagination FIXTHIS %}

    **example**::

        {% bootstrap_pagination FIXTHIS %}
    """

    pagination_kwargs = {'extra': 'search=%s' % search, 'page': page}
    return get_pagination_context(**pagination_kwargs)


class CurlNode(template.Node):

    def __init__(self, repo=None):
        super().__init__()
        self.repo_token = repo

    def render(self, context):
        if self.repo_token and not context[self.repo_token].is_private:
            return mark_safe('curl')
        if context['df_remote_authenticated']:
            return mark_safe('curl --anyauth -u :')
        return mark_safe('curl --basic -u $MONETA_USER:$MONETA_PASSWORD')


# noinspection PyUnusedLocal
@register.tag('curl')
def do_curl(parser, token):
    bits = list(token.split_contents())
    tagname = bits[0]
    repo = bits[1] if len(bits) > 1 else None
    return CurlNode(repo)



# coding=utf-8
import re

__author__ = 'flanker'

"""Setup file for the Moneta project.
"""

import codecs
import os.path

from setuptools import setup, find_packages

with codecs.open(os.path.join(os.path.dirname(__file__), 'README.rst'), encoding='utf-8') as fd:
    long_description = fd.read()
version = None
for line in codecs.open(os.path.join('moneta', '__init__.py'), 'r', encoding='utf-8'):
    matcher = re.match(r"""^__version__\s*=\s*['"](.*)['"]\s*$""", line)
    version = version or matcher and matcher.group(1)

entry_points = {'console_scripts': ['moneta-manage = djangofloor.scripts:manage',
                                    'moneta-gunicorn = djangofloor.scripts:gunicorn']}

setup(
    name='moneta',
    version=version,
    description='Emulator for different kinds (Debian/Yum/Pypi/Maven/Ruby) of package repositories on the same server.',
    long_description=long_description,
    author='flanker',
    author_email='flanker@19pouces.net',
    license='CeCILL-B',
    url='',
    entry_points=entry_points,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=['setuptools>=1.0', 'djangofloor>=0.17', 'python-gnupg', 'rubymarshal', 'pyyaml'],
    setup_requires=[],
    classifiers=[],
)



# -*- coding: utf-8 -*-
import pkg_resources
from moneta.repositories.aptitude import Aptitude
from moneta.repositories.tests import RepositoryTestCase

__author__ = 'flanker'


class TestYumRpm(RepositoryTestCase):

    def test_add_file(self):
        repo = self.create_repository(Aptitude)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', 'lib3ds-dev_1.3.0-8_mips.deb')
        self.add_file_to_repository(repo, filename)

    def test_generate_index(self):
        repo = self.create_repository(Aptitude)
        filename = pkg_resources.resource_filename('moneta.repositories.tests', 'lib3ds-dev_1.3.0-8_mips.deb')
        self.add_file_to_repository(repo, filename)
        aptitude = Aptitude()
        aptitude.generate_indexes(repo)



# -*- coding: utf-8 -*-
from django.http import HttpRequest
from django.test import TestCase
from moneta.repositories.maven3 import Maven3
from moneta.repository.models import Repository, Element

__author__ = 'flanker'


class TestMaven3(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.repo = Repository(archive_type=Maven3.archive_type, name='test_maven3')
        cls.repo.save()
        cls.elts = [
            Element(version='1.1.4c',
                    archive='xpp3.xpp3_min',
                    name='xpp3_min',
                    filename='xpp3_min-1.1.4c.jar',
                    full_name='xpp3_min-1.1.4c.jar',
                    sha1='19d4e90b43059058f6e056f794f0ea4030d60b86', ),
            Element(version='1.0-2',
                    archive='javax.xml.stream.stax-api',
                    name='stax-api',
                    filename='stax-api-1.0-2.jar',
                    full_name='stax-api-1.0-2.jar',
                    sha1='d6337b0de8b25e53e81b922352fbea9f9f57ba0b', ),
            Element(version='1.0-2',
                    archive='javax.xml.stream.stax-api',
                    name='stax-api',
                    filename='stax-api-1.0-2.pom',
                    full_name='stax-api-1.0-2.pom',
                    sha1='5379b69f557c5ab7c144d22bf7c3768bd2adb93d', ),
            Element(version='2.2.2',
                    archive='javax.xml.bind.jaxb-api',
                    name='jaxb-api',
                    filename='jaxb-api-2.2.2.jar',
                    full_name='jaxb-api-2.2.2.jar',
                    sha1='aeb3021ca93dde265796d82015beecdcff95bf09', ),
            Element(version='2.2.2',
                    archive='javax.xml.bind.jaxb-api',
                    name='jaxb-api',
                    filename='jaxb-api-2.2.2.pom',
                    full_name='jaxb-api-2.2.2.pom',
                    sha1='a8368234f7555dd64d3a9060a0b02e6c215694fb', ),
        ]
        for elt in cls.elts:
            elt.repository = cls.repo
        Element.objects.bulk_create(cls.elts)

    def test_count(self):
        self.assertEqual(Element.objects.filter(repository=self.repo).count(), 5)

    def test_xpp3(self):
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3')), 1)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min')), 1)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_m')), 0)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/1.1.4')), 0)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/')), 1)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/')), 1)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/1.1.4')), 0)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/1.1.4c')), 1)
        self.assertEqual(len(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/1.1.4c/')), 1)
        self.assertIsInstance(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/1.1.4c/xpp3_min-1.1.4c.jar'), Element)
        self.assertIsInstance(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/1.1.4c/xpp3_min-1.1.4c.jar.sha1'), str)
        self.assertEqual(Maven3.browse_repo_inner(self.repo.id, 'xpp3/xpp3_min/1.1.4c/xpp3_min-1.1.4c.jar.sha1'),
                         '19d4e90b43059058f6e056f794f0ea4030d60b86')
        self.assertEqual(Element.objects.filter(repository=self.repo).count(), 5)

    def test_javax(self):
        result = Maven3.browse_repo_inner(self.repo.id, 'javax')
        self.assertEqual(len(result), 1)
        self.assertIsInstance(result, dict)
        self.assertTrue('javax' in result)

    def test_all(self):
        result = Maven3.browse_repo_inner(self.repo.id, '')
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result, dict)
        self.assertTrue('javax' in result)
        self.assertTrue('xpp3' in result)



# coding=utf-8
from django import forms
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils.text import slugify
from django.utils.translation import ugettext as _
import functools
from moneta.repository.models import Repository, ElementSignature

__author__ = 'flanker'


def repo_name_validator(value):
    if Repository.objects.filter(slug=slugify(value)).count() > 0:
        raise ValidationError(_('This repository already exists.'))


@functools.lru_cache()
def get_repository_form():
    from moneta.repositories.base import RepositoryModelsClasses

    class RepositoryForm(forms.Form):
        name = forms.CharField(max_length=80, label=_('Repository name'), validators=[repo_name_validator])
        archive_type = forms.ChoiceField(label=_('Type of archives'), choices=RepositoryModelsClasses())
        on_index = forms.BooleanField(label=_('Add on public index?'), initial=True, required=False)
        is_private = forms.BooleanField(label=_('Authentication required?'), initial=False, required=False, widget=forms.HiddenInput())
        states = forms.CharField(label=_('Possible states for archives'), initial='qualif prod',
                                 help_text=_('Please separate values by spaces'),
                                 validators=[RegexValidator('\w+(\s\w)*')])
        admin_group = forms.ModelMultipleChoiceField(Group.objects.all(), label=_('Groups allowed to upload packages'),
                                                     required=False)
        reader_group = forms.ModelMultipleChoiceField(Group.objects.all(), widget=forms.HiddenInput(),
                                                      label=_('Groups allowed to download packages'),
                                                      help_text=_('Only if downloads require authentication'), required=False)

    return RepositoryForm


CONFIRM_PHRASE = _('yes, I want to delete this element.')


def confirm_phrase(value):
    if value != CONFIRM_PHRASE:
        raise ValidationError(_('You must confirm the deletion.'))


class DeleteRepositoryForm(forms.Form):

    confirm = forms.CharField(max_length=80, label=_('Please enter “%(confirm)s”') % {'confirm': CONFIRM_PHRASE},
                              validators=[confirm_phrase])


class RepositoryUpdateForm(forms.Form):
    on_index = forms.BooleanField(label=_('Display on index for everybody?'), initial=True, required=False)
    is_private = forms.BooleanField(label=_('Must downloads be authenticated?'), initial=False, required=False, widget=forms.HiddenInput())
    states = forms.CharField(label=_('Possible states'), initial='dev qualif prod',
                             help_text=_('Please separate values by spaces'),
                             validators=[RegexValidator('\w+(\s\w)*')])
    admin_group = forms.ModelMultipleChoiceField(Group.objects.all(), label=_('Groups allowed to upload'),
                                                 required=False)
    # reader_group = forms.ModelMultipleChoiceField(Group.objects.all(), widget=forms.HiddenInput(),
    #                                               help_text=_('Only if downloads are authenticated'),
    #                                               label=_('Groups allowed to download packages'), required=False)


class SignatureForm(forms.Form):
    sha256 = forms.CharField(label=_('SHA256'), max_length=255, validators=[RegexValidator(r'\w{64}')])
    method = forms.ChoiceField(label=_('Method'), choices=ElementSignature.METHODS)



# -*- coding: utf-8 -*-
from djangofloor.utils import DirectoryPath

__author__ = 'flanker'

FLOOR_URL_CONF = 'moneta.root_urls.urls'
EXTRA_INSTALLED_APP = 'bootstrap3'
FLOOR_INSTALLED_APPS = ['moneta', 'moneta.repositories', 'moneta.repository', '{EXTRA_INSTALLED_APP}']
FLOOR_INDEX = 'moneta.views.index'
FLOOR_PROJECT_NAME = 'Moneta'

UPLOAD_ROOT = 'uploads'
ARCHIVE_FILTERS = [
    'moneta.repository.filters.informations',
    # 'moneta.repository.filters.deb_archive',
]
REPOSITORY_CLASSES = [
    'moneta.repositories.aptitude.Aptitude',
    'moneta.repositories.pypi.Pypi',
    'moneta.repositories.maven3.Maven3',
    'moneta.repositories.flat_files.FlatFile',
    'moneta.repositories.yum.Yum',
    'moneta.repositories.ruby.RubyGem',
    'moneta.repositories.jetbrains.Jetbrains',
]

TEMPLATE_CONTEXT_PROCESSORS = [
    'django.contrib.auth.context_processors.auth',
    'django.core.context_processors.debug',
    'django.core.context_processors.request',
    'django.core.context_processors.i18n',
    'django.core.context_processors.media',
    'django.core.context_processors.static',
    'django.core.context_processors.tz',
    'django.contrib.messages.context_processors.messages',
    'djangofloor.context_processors.context_base',
    'moneta.context_processors.context_base',
]
TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                # Already defined Django-related contexts here
                'django.contrib.auth.context_processors.auth',
                'django.core.context_processors.debug',
                'django.core.context_processors.request',
                'django.core.context_processors.i18n',
                'django.core.context_processors.media',
                'django.core.context_processors.static',
                'django.core.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
                'djangofloor.context_processors.context_base',
                'moneta.context_processors.context_base',
                # `allauth` needs this from django
                'django.template.context_processors.request',
            ],
        },
    },
]

STORAGES = {
    'archive': {
        'ENGINE': 'moneta.repository.storages.FlatStorage',
        'ROOT': DirectoryPath('{LOCAL_PATH}/storage/archives'),
        'PATH_LEN': 1,
    },
    'default': {
        'ENGINE': 'moneta.repository.storages.FlatStorage',
        'ROOT': DirectoryPath('{LOCAL_PATH}/storage/uncompressed'),
        'PATH_LEN': 1,
    },
    'cache': {
        'ENGINE': 'moneta.repository.storages.FlatStorage',
        'ROOT': DirectoryPath('{LOCAL_PATH}/storage/cache'),
        'PATH_LEN': 1,
    }
}

STORAGE_ARCHIVE = 'archive'
STORAGE_UNCOMPRESSED = 'uncompressed'
STORAGE_CACHE = 'cache'
CACHES = {'default': {'BACKEND': 'django.core.cache.backends.dummy.DummyCache'}}

FLOOR_FAKE_AUTHENTICATION_USERNAME = None
DEBUG = True
# TO BE CONFIGURED
GNUPG_HOME = DirectoryPath('{LOCAL_PATH}/gpg')
# TO BE CONFIGURED
TEMP_ROOT = DirectoryPath('{LOCAL_PATH}/tmp')
# TO BE CONFIGURED
GNUPG_KEYID = '1DA759EA7F5EF06F'
# TO BE CONFIGURED
GNUPG_PATH = 'gpg'
